package org.eclipse.palamedes.kif.core.ast;

import org.eclipse.palamedes.kif.core.ast.visitor.NotationWriter;


public abstract class ASTNode {

    private final AST       ast;

    public static final int CONJUNCTION  = 1;

    public static final int DISJUNCTION  = 2;

    public static final int FORM         = 3;

    public static final int FUNTERM      = 4;

    public static final int HORNSENT     = 5;

    public static final int IMPLICATION  = 6;

    public static final int INDVAR       = 7;

    public static final int LOGCONST     = 8;

    public static final int NEGATION     = 10;

    public static final int OBJCONST     = 11;

    public static final int OPERATOR     = 12;

    public static final int RELSENT      = 13;

    public static final int SEQUENCE     = 15;

    public static final int COMMENT      = 16;


    /**
     * A character index into the original source string, or <code>-1</code>
     * if no source position information is available for this node;
     * <code>-1</code> by default.
     */
    private int             sourceStart  = -1;

    /**
     * A character length, or <code>0</code> if no source position information
     * is recorded for this node; <code>0</code> by default.
     */
    private int             sourceLength = 0;

    /**
     * Integer containing the node type in the top 16 bits and flags in the
     * bottom 16 bits; none set by default.
     * <p>
     * N.B. This is a private field, but declared as package-visible for more
     * efficient access from inner classes.
     * </p>
     *
     */
    int                     typeAndFlags = 0;

    ASTNode(AST ast) {

        if (ast == null) {
            throw new IllegalArgumentException();
        }

        this.ast = ast;
        this.setNodeType(getNodeType0());

    }

    public void setSourceStart(int sourceStart) {
        this.sourceStart = sourceStart;
    }

    public void setSourceLength(int sourceLength) {
        this.sourceLength = sourceLength;
    }

    public final int getSourceStart() {
        return sourceStart;
    }

    public final int getSourceLength() {
        return sourceLength;
    }

    public final AST getAST() {
        return this.ast;
    }

    /**
     * Returns an integer value identifying the type of this concrete AST node.
     * The values are small positive integers, suitable for use in switch
     * statements.
     * <p>
     * For each concrete node type there is a unique node type constant (name
     * and value). The unique node type constant for a concrete node type such
     * as <code>CastExpression</code> is <code>ASTNode.CAST_EXPRESSION</code>.
     * </p>
     *
     * @return one of the node type constants
     */
    public final int getNodeType() {
        return this.typeAndFlags >>> 16;
    }

    /**
     * Returns an integer value identifying the type of this concrete AST node.
     * <p>
     * This internal method is implemented in each of the concrete node
     * subclasses.
     * </p>
     *
     * @return one of the node type constants
     */
    abstract int getNodeType0();

    public final ASTNode clone(AST baseAST) {
        return clone0(ast);
    }

    abstract ASTNode clone0(AST baseAST);

    /**
     * Sets the integer value identifying the type of this concrete AST node.
     * The values are small positive integers, suitable for use in switch
     * statements.
     *
     * @param nodeType
     *            one of the node type constants
     */
    private void setNodeType(int nodeType) {
        int old = this.typeAndFlags & 0xFFFF0000;
        this.typeAndFlags = old | (nodeType << 16);
    }

    /**
     * The <code>ASTNode</code> implementation of this <code>Object</code>
     * method uses object identity (==). Use <code>subtreeMatch</code> to
     * compare two subtrees for equality.
     *
     * @param obj
     *            {@inheritDoc}
     * @return {@inheritDoc}
     */
    @Override
    public final boolean equals(Object obj) {
        return this == obj; // equivalent to Object.equals
    }

    /*
     * (non-Javadoc) This makes it consistent with the fact that a equals
     * methods has been provided.
     *
     * @see java.lang.Object#hashCode()
     */
    @Override
    public final int hashCode() {
        return super.hashCode();
    }

    /**
     * Returns whether the subtree rooted at the given node matches the given
     * other object as decided by the given matcher.
     * <p>
     * This internal method is implemented in each of the concrete node
     * subclasses.
     * </p>
     *
     * @param matcher
     *            the matcher
     * @param other
     *            the other object, or <code>null</code>
     * @return <code>true</code> if the subtree matches, or <code>false</code>
     *         if they do not match
     */
    abstract boolean subtreeMatch0(ASTMatcher matcher, Object other);


    /**
     * This method is for internal use!
     *
     * It is assumed that the current node contains no own information and
     * should be a clone of the given node. If you are looking for a
     * copy/clone method to get a deep copy of a ASTNode use clone() instead.
     *
     * @see clone()
     */
    protected final void copyMembers(ASTNode node) {
        sourceLength = node.sourceLength;
        sourceStart  = node.sourceStart;
    }

    public abstract void traverse(ASTVisitor visitor);

    
	/**
	 * @return the prefix form of this term
	 * e.g., "f(a,g(1,2))" 
	 */
	public final String getPrefixNotation(){
		return new NotationWriter(NotationWriter.PREFIX).toString(this);
	}

	/**
	 * @return the infix KIF form of this term
	 * e.g., "(f a (g 1 2))" 
	 */
	public final String getInfixNotation(){
		return new NotationWriter(NotationWriter.INFIX).toString(this);
	}

	@Override
	public String toString() {
		return getInfixNotation();
	}

}
