package org.eclipse.palamedes.kif.core.ast.visitor;

import org.eclipse.palamedes.kif.core.ast.ASTNode;
import org.eclipse.palamedes.kif.core.ast.ASTVisitor;
import org.eclipse.palamedes.kif.core.ast.KIFConjunction;
import org.eclipse.palamedes.kif.core.ast.KIFDisjunction;
import org.eclipse.palamedes.kif.core.ast.KIFForm;
import org.eclipse.palamedes.kif.core.ast.KIFFunTerm;
import org.eclipse.palamedes.kif.core.ast.KIFImplication;
import org.eclipse.palamedes.kif.core.ast.KIFIndVar;
import org.eclipse.palamedes.kif.core.ast.KIFLogConst;
import org.eclipse.palamedes.kif.core.ast.KIFNegation;
import org.eclipse.palamedes.kif.core.ast.KIFObjConst;
import org.eclipse.palamedes.kif.core.ast.KIFOperator;
import org.eclipse.palamedes.kif.core.ast.KIFRelSent;

public class DefaultVisitor extends ASTVisitor {

    public static final boolean IMP_HEAD_FIRST = true;
    public static final boolean IMP_HEAD_LAST  = false;

    private boolean headFirst = IMP_HEAD_LAST;

    protected ASTNode currentTopLevelNode = null;
    protected boolean insideImplication   = false;
    protected boolean insideHead		  = false;
    protected boolean insideBody		  = false;

    @Override
    public void visit(KIFConjunction node) {
        preVisitComplexNode(node);
        for (ASTNode childNode : node.getOperands() )
            childNode.traverse(this);
        postVisitComplexNode(node);
    }

    @Override
    public void visit(KIFDisjunction node) {
        preVisitComplexNode(node);
        for (ASTNode childNode : node.getOperands() )
            childNode.traverse(this);
        postVisitComplexNode(node);
    }

    @Override
    public void visit(KIFForm node) {

        // there was no form, so we can not do anything
        if (node == null)
            return;

        preVisitFormNode(node);
        for (ASTNode childNode : node.getSentences() ) {
            currentTopLevelNode = childNode;
            childNode.traverse(this);
        }
        postVisitFormNode(node);
    }

    @Override
    public void visit(KIFFunTerm node) {
        preVisitComplexNode(node);
        for (ASTNode childNode : node.getTerms() )
            childNode.traverse(this);
        postVisitComplexNode(node);
    }

    @Override
    public void visit(KIFImplication node) {
        if (headFirst) {
            preVisitComplexNode(node);
            insideImplication = true;
            insideHead		  = true;

            // and then the consequence
            node.getConsequence().traverse(this);

            insideHead		  = false;
            insideBody		  = true;

            // first we traverse the premises
            for (ASTNode childNode : node.getPremise() )
                childNode.traverse(this);

            insideBody        = false;
            insideImplication = false;
            postVisitComplexNode(node);
        } else  {
            preVisitComplexNode(node);
            insideImplication = true;
            insideBody		  = true;

            // first we traverse the premises
            for (ASTNode childNode : node.getPremise() )
                childNode.traverse(this);

            insideBody        = false;
            insideHead		  = true;

            // and then the consequence
            node.getConsequence().traverse(this);

            insideHead		  = false;
            insideImplication = false;
            postVisitComplexNode(node);
        }
    }

    @Override
    public void visit(KIFIndVar node) {
        visitBasicNode(node);
    }

    @Override
    public void visit(KIFLogConst node) {
        visitBasicNode(node);
    }

    @Override
    public void visit(KIFNegation node) {
        preVisitComplexNode(node);
        for (ASTNode childNode : node.getOperands() )
            childNode.traverse(this);
        postVisitComplexNode(node);
    }

    @Override
    public void visit(KIFObjConst node) {
        visitBasicNode(node);
    }

    @Override
    public void visit(KIFOperator node) {
        visitBasicNode(node);
    }

    @Override
    public void visit(KIFRelSent node) {
        preVisitComplexNode(node);
        for (ASTNode childNode : node.getTerms() )
            childNode.traverse(this);
        postVisitComplexNode(node);
    }

    /*
     * Helper hooks that should make it simpler to write own AST visitors.
     */
    protected void visitBasicNode		(ASTNode node) {}
    protected void postVisitComplexNode (ASTNode node) {}
    protected void preVisitComplexNode  (ASTNode node) {}
    protected void postVisitFormNode    (ASTNode node) {}
    protected void preVisitFormNode     (ASTNode node) {}

    public boolean isHeadFirst() {
        return headFirst;
    }

    public void setHeadFirst(boolean headFirst) {
        this.headFirst = headFirst;
    }
}
