package org.eclipse.palamedes.kif.core.knowledgebase;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.TreeSet;

import org.eclipse.palamedes.kif.core.ast.ASTNode;
import org.eclipse.palamedes.kif.core.ast.KIFImplication;
import org.eclipse.palamedes.kif.core.ast.KIFSent;

/**
 * Base class for the knowledge base entries.
 */
public class KBEntry {

    /** Property name for name */
    public static final String P_NAME = "Name";

    /** Property name for type */
    public static final String P_TYPE = "Type";

    /** Property name for arity */
    public static final String P_ARITY = "Arity";
    
    /** Property name for <...> */
    public static final String P_TOP_LEVEL_NODES = "Top Level Nodes";

    /** Property name for <...> */
    public static final String P_IMP_HEAD = "Implication Head";

    /** Property name for <...> */
    public static final String P_IMP_BODY = "Implication Body";
    
    /** Property name for <...> */
    public static final String P_NODES = "Node List";
    
    
    /** Property value relation */
    public static final String V_REL = "Relation";

    /** Property value function */
    public static final String V_FUN = "Function";

    final HashMap<String, Object> properties = new HashMap<String, Object>();

    public KBEntry(String name) {
    	properties.put( P_NAME, 		   name );
    	properties.put( P_TOP_LEVEL_NODES, new ArrayList<KIFSent>() );
    	properties.put( P_IMP_HEAD, 	   new ArrayList<KIFImplication>() );
    	properties.put( P_IMP_BODY, 	   new ArrayList<KIFImplication>() );
    	properties.put( P_NODES, 	   	   new ArrayList<ASTNode>() );
    }

    
    /**
     * This is a convenience method which returns true if the propertyName 
     * exists; otherwise false. 
     */
    public boolean has(String propertyName) {
        return properties.containsKey(propertyName);
    }


    /**
     * Returns the value of a property given by propertyName. 
     */
    public Object get(String propteryName) {
    	return properties.get(propteryName);
    }
    
    
    /**
     * Sets the property propertyName to the value propertyValue. If the map 
     * previously contained a mapping for the key, the old value is replaced.
     * @return the previous value associated with key, or null if there was no 
     * 		   mapping for key. (A null return can also indicate that the map 
     * 		   previously associated null with key.) 
     */
    public Object put(String propteryName, Object propertyValue) {
    	return properties.put(propteryName, propertyValue);
    }
    
    /**
     * Returns the property names.
     */
    public String[] getPropertyNames() {
    	TreeSet<String> keys = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
    	keys.addAll(properties.keySet());
    	return keys.toArray( new String[properties.size()] );
    }
    
    /**
     * Returns the property names but only thus which are not in toBeFiltered.
     */
    public String[] getPropertyNames(String[] toBeFiltered) {
    	TreeSet<String> keys = new TreeSet<String>(String.CASE_INSENSITIVE_ORDER);
    	keys.addAll(properties.keySet());
    	for (String str : toBeFiltered)
    		keys.remove(str);
    	return keys.toArray( new String[keys.size()] );
    }

    /**
     * Returns the name of the entry.
     */
    public String getName() {
        return (String) properties.get(P_NAME);
    }

    /**
     * Returns all ASTNodes that correspond to the symbol directly.
     */
    @SuppressWarnings("unchecked")
	public ArrayList<ASTNode> getNodeList() {
        return (ArrayList<ASTNode>) properties.get(P_NODES);
    }


    /**
     * Returns all Top Level ASTNodes corresponding to the symbol.
     */
    @SuppressWarnings("unchecked")
	public ArrayList<KIFSent> getTopLevelNodes() {
        return  (ArrayList<KIFSent>) properties.get(P_TOP_LEVEL_NODES);
    }


    /**
     * Returns all implications where the symbol appears in the head.
     */
    @SuppressWarnings("unchecked")
	public ArrayList<KIFImplication> getImplicationHead() {
        return  (ArrayList<KIFImplication>) properties.get(P_IMP_HEAD);
    }


    /**
     * Returns all implications where the symbol appears in the body.
     */
    @SuppressWarnings("unchecked")
	public ArrayList<KIFImplication> getImplicationBody() {
        return  (ArrayList<KIFImplication>) properties.get(P_IMP_BODY);
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder();

        sb.append("\nProperties:");
        for (Object propertyName : getPropertyNames() ) {
        	sb.append("\n\t").append(propertyName);
        	sb.append(":\t").append(properties.get(propertyName));
        }
        
        return sb.toString();
    }
}
