package org.eclipse.palamedes.kif.core.knowledgebase;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Arrays;
import java.util.TreeMap;

import org.eclipse.palamedes.kif.core.ast.KIFForm;
import org.eclipse.palamedes.kif.core.ast.KIFFunTerm;
import org.eclipse.palamedes.kif.core.ast.KIFImplication;
import org.eclipse.palamedes.kif.core.ast.KIFLogConst;
import org.eclipse.palamedes.kif.core.ast.KIFRelSent;
import org.eclipse.palamedes.kif.core.ast.visitor.DefaultVisitor;


class KnowledgeBaseVisitor extends DefaultVisitor {

    private final TreeMap<String, KBEntry> kbEntries;
    private boolean                        hasError;

    KnowledgeBaseVisitor(KIFForm form) {
        kbEntries = new TreeMap<String, KBEntry>(String.CASE_INSENSITIVE_ORDER);
        visit(form);
    }

    public boolean hasError() {
        return hasError;
    }

    public TreeMap<String, KBEntry> getEntries() {
        return kbEntries;
    }

    /***************************************************************************
     *
     * Information gathering through AST visitor pattern
     * @throws Exception
     *
     **************************************************************************/
    @Override
    public void visit(KIFRelSent node) {

        String name = node.getName();

        // create key
        if ( !kbEntries.containsKey(name) )
            kbEntries.put(name, new KBEntry(name) );

        // get description object
        KBEntry info = kbEntries.get(name);
        info.put(KBEntry.P_TYPE, KBEntry.V_REL);

        // set arity property
        try {
            setArity(info, node.getTerms().size());
        } catch (Exception exp) {
            hasError = true;
            info.put("Exception", getExceptionString(exp));
        }

        // we add the node to the node_list
        info.getNodeList().add(node);

        // we are on the top
        if (currentTopLevelNode == node) {
            info.getTopLevelNodes().add(node);

        // we are within an implication
        } else if (currentTopLevelNode instanceof KIFImplication) {
            KIFImplication implication = (KIFImplication) currentTopLevelNode;

            if (insideHead)
                info.getImplicationHead().add(implication);
            else
                info.getImplicationBody().add(implication);
        }
        super.visit(node);

    }

    @Override
    public void visit(KIFFunTerm node) {
        String name = node.getName();

        // create key
        if ( !kbEntries.containsKey(name) )
            kbEntries.put(name, new KBEntry(name) );

        // get description object
        KBEntry info = kbEntries.get(name);
        info.put(KBEntry.P_TYPE, KBEntry.V_FUN);

        // set arity property
        try {
            setArity(info, node.getTerms().size());
        } catch (Exception exp) {
            hasError = true;
            info.put("Exception", getExceptionString(exp));
        }

        // we add the node to the node_list
        info.getNodeList().add(node);

        // we are on the top
        if (currentTopLevelNode == node) {
            System.err.println("Should not be a top level node: " + node);

        // we are within an implication
        } else if (currentTopLevelNode instanceof KIFImplication) {
            KIFImplication implication = (KIFImplication) currentTopLevelNode;

            if (insideHead)
                info.getImplicationHead().add(implication);
            else
                info.getImplicationBody().add(implication);
        }
        super.visit(node);
    }

    /**
     * This method is a hack to get also the relations without arguments into
     * the relation map.
     */
    @Override
    public void visit(KIFLogConst node) {

        String name = node.getName();

        // create key
        if ( !kbEntries.containsKey(name) )
            kbEntries.put(name, new KBEntry(name) );

        // get description object
        KBEntry info = kbEntries.get(name);
        info.put(KBEntry.P_TYPE, KBEntry.V_REL);

        // set arity property
        setArity(info, 0);

        // we add the node to the node_list
        info.getNodeList().add(node);

        // we are on the top
        if (currentTopLevelNode == node) {
            info.getTopLevelNodes().add(node);

        // we are within an implication
        } else if (currentTopLevelNode instanceof KIFImplication) {
            KIFImplication implication = (KIFImplication) currentTopLevelNode;

            if (insideHead)
                info.getImplicationHead().add(implication);
            else
                info.getImplicationBody().add(implication);
        }
        super.visit(node);

    }

    /**
     * This method sets the arity property in a description object.
     * @param info
     * @param arity
     */
     private static void setArity(KBEntry info, int arity) {

         String arityString = (String) info.get(KBEntry.P_ARITY);

         // we do not have anything yet
         if ( arityString == null ) {
             info.put(KBEntry.P_ARITY, arity + "");
             return;
         }

         // retrieve existing arity information
         String[] listStr = arityString.split(", ");
         int[]    listInt = new int[listStr.length];
         for (int i = 0; i < listStr.length; i++) {
             listInt[i] = Integer.parseInt(listStr[i]);
         }

         // we only change something if we got a new arity
         if ( Arrays.binarySearch(listInt, arity) == -1) {

             // create a new list, add the new value and sort it
             int[] newList = new int[listInt.length + 1];
             for (int i = 0; i < listInt.length; i++) {
                 newList[i] = listInt[i];
             }
             newList[listInt.length] = arity;
             Arrays.sort(newList);

             // to value
             StringBuffer sb = new StringBuffer();
             for (int value : newList)
                 sb.append(value).append(", ");
             arityString = sb.toString();

             info.put("hasMultipleArity", true);
         }

         info.put(KBEntry.P_ARITY, arityString);
     }

    public static String getExceptionString(Exception exp) {
        StringWriter sw = new StringWriter();
        PrintWriter  pw = new PrintWriter(sw, true);
        exp.printStackTrace(pw);
        pw.flush();
        sw.flush();
        return sw.toString();
    }
}
