package org.eclipse.palamedes.kif.core.knowledgebase.analyser;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.BitSet;
import java.util.HashSet;
import java.util.List;

import org.eclipse.palamedes.kif.core.ast.KIFObjConst;
import org.eclipse.palamedes.kif.core.knowledgebase.IKBAnalyser;
import org.eclipse.palamedes.kif.core.knowledgebase.IKBFilter;
import org.eclipse.palamedes.kif.core.knowledgebase.KBEntry;
import org.eclipse.palamedes.kif.core.knowledgebase.KnowledgeBase;


public class ABinaryFacts implements IKBAnalyser {

    // we only care about binary facts with resolved domains
    static class ABFFilter implements IKBFilter {

        public boolean isMatching(KBEntry entry) {
            return "2".equals( entry.get(KBEntry.P_ARITY) ) &&
                   entry.has("isFact")                      &&
                   entry.has("domain_distinct")             &&
                   entry.has("domain_0")                    &&
                   entry.has("domain_1");
        }

    }

    // Temporary created or saved values
    private int[][] 		testArray;
    private int[]			domainElements;
    private HashSet<String> domainDistinct;


    public void analyse(KnowledgeBase kb) {
        for (KBEntry entry : kb.filter(new ABFFilter()))
            testRelationProperties(entry);
    }

    @SuppressWarnings("unchecked")
    private void testRelationProperties(KBEntry entry) {

        // get information from properties
        Object rawDomainDistinct = entry.get("domain_distinct");
        Object rawDomain0 		 = entry.get("domain_0");
        Object rawDomain1 		 = entry.get("domain_1");

        // convert it back
        domainDistinct 			  = (HashSet)           rawDomainDistinct;
        List<KIFObjConst> domain0 = (List<KIFObjConst>) rawDomain0;
        List<KIFObjConst> domain1 = (List<KIFObjConst>) rawDomain1;

        // create test arrays: convert lower case names to hash code
        testArray  	   = new int[domain0.size()][];
        domainElements = new int[domainDistinct.size()];
        String name0, name1;
        for (int i = 0; i < testArray.length; i++) {
            name0 = (domain0.get(i)).getName().toLowerCase();
            name1 = (domain1.get(i)).getName().toLowerCase();
            testArray[i] = new int[] { name0.hashCode(), name1.hashCode() };
        }

        int i = 0;
        for (Object element : domainDistinct) {
            name0 = ((String) element).toLowerCase();
            domainElements[i++] = name0.hashCode();
        }

        // run the tests
        isReflexive(entry);
        isIrreflexive(entry);
        isSymmetric(entry);
        isAsymmetric(entry);
        isAntisymmetric(entry);
        // TODO: isTransitiv(info); -> do we need to check this ?

        // find special stuff
        findEndElements(entry);
        findOrderedRelation(entry);
    }


    /**
     * This method tests if the relation is reflexive.
     * That means: for all elements x a tuple(x,x) has to be in the relation.
     *
     * Complexity: O(n²) with n = domain.length
     *
     * @param info
     */
    private void isReflexive(KBEntry info) {

        // check for each element
        int[]   testTuple;
        boolean found;
        for (int element : domainElements) {

            // get tuple and create reflexive versions
            testTuple = new int[] { element, element };

            // we assume that we did not found a reflexive element so far
            found = false;

            // compare to all other tuples
            for ( int j = 0; j < testArray.length; j++ ) {

                // found one ? -> stop lookup !
                if ( Arrays.equals(testTuple, testArray[j]) ) {
                    found = true;
                    break;
                }
            }

            // !(x,x) element of relation => not reflexive
            // failed -> exit here
            if ( !(found) )
                return;
        }

        // set property
        info.put("isReflexive", true);

    }

    /**
     * This method tests if the relation is irreflexive.
     * That means: for no element x a tuple(x,x) has to be in the relation
     *
     * Complexity: O(n) with n = domain.size
     *
     * @param info
     */
    private void isIrreflexive(KBEntry entry) {

        for (int i = 0; i < testArray.length; i++) {

            // if there is a tuple (x,x) then leave this test
            if (testArray[i][0] == testArray[i][1])

                // failed -> exit here
                return;
        }

        // we survived ? so it is irreflexive
        entry.put("isIrreflexive", true);
    }


    /**
     * This method tests if the relation is symmetric.
     * That means: if relation contains a tuple (x,y) it also has to have a
     *             tuple (y,x)
     *
     * Complexity: O( (n²-1)/2 ) with n = domain.size
     *
     * @param info
     */
    private void isSymmetric(KBEntry entry) {

        BitSet testSet = new BitSet(testArray.length);
        testSet.set( 0, testArray.length + 1);

        int[]   tmpTuple;
        int[]   testTuple;
        boolean found;
        for (int i = 0; i < testArray.length; i++ ) {

            if ( testSet.get(i) ) {
                testSet.clear(i);

                // get tuple and swap position to get the reversed tuple
                tmpTuple    = testArray[i];
                testTuple   = new int[] { tmpTuple[1], tmpTuple[0] };
                found       = false;

                // compare to all other following tuples
                // reflexive elements are tested as well through j = i
                // instead of j = i + 1
                for ( int j = i; j < testArray.length; j++ ) {
                    if ( Arrays.equals( testTuple, testArray[j] ) ) {
                        found = true;
                        testSet.clear(j);
                        break;
                    }
                }

                // failed -> exit here
                if ( !(found) )
                    return;
            }
        }

        entry.put("isSymmetric", true);
    }


    /**
     * This method tests if the relation is asymmetric.
     * That means: for no tuple(x,y) a tuple(y,x) exists in the relation
     *
     * Complexity: O( (n²-1)/2 ) with n = domain.size
     *
     * @param info
     */
    private void isAsymmetric(KBEntry entry) {

        if ( entry.has("isReflexive") )
            return;

        int[] tmpTuple;
        int[] testTuple;
        for (int i = 0; i < testArray.length; i++ ) {

            // get tuple and swap position to get the reversed tuple
            tmpTuple    = testArray[i];
            testTuple   = new int[] { tmpTuple[1], tmpTuple[0] };

            // compare to all other following tuples
            for ( int j = i; j < testArray.length; j++ )

                // failed -> exit here
                if ( Arrays.equals( testTuple, testArray[j] ) )
                    return;
        }

        // we survived ? so it is asymmetric
        entry.put("isAsymmetric", true);
    }

    /**
     * This method tests if the relation is antisymmetric.
     * That means: the relation is neither symmetric nor antisymmetric
     *
     * @param info
     */
    private void isAntisymmetric(KBEntry entry) {

        if ( entry.has("isAsymmetric") ||
             entry.has("isSymmetric")   )
            return;

        // we survived ? so it is antisymmetric
        entry.put("isAntisymmetric", true);
    }

    /**
     * This method looks for elements that appear only in one position. These
     * elements could be bottom/top elements of a ordered relation.
     *
     * Complexity: O(n) with n = domain.size
     * @param info
     */
    private void findEndElements(KBEntry entry) {

        List<Integer> endList0 = new ArrayList<Integer>();
        List<Integer> endList1 = new ArrayList<Integer>();
        for (int element : domainElements) {
            endList0.add(element);
            endList1.add(element);
        }

        for (int i = 0; i < testArray.length; i++) {
            if (endList0.indexOf(testArray[i][0]) > -1)
                endList0.remove( endList0.indexOf(testArray[i][0]) );
            if (endList1.indexOf(testArray[i][1]) > -1)
                endList1.remove( endList1.indexOf(testArray[i][1]) );
        }

        // find the names for the hash codes
        ArrayList<String> endStrings0 = new ArrayList<String>();
        ArrayList<String> endStrings1 = new ArrayList<String>();

        for ( int hash : endList0 ) {
            for (String name : domainDistinct)
                if (name.hashCode() == hash)
                    endStrings0.add(name);
        }

        for ( int hash : endList1 ) {
            for (String name : domainDistinct)
                if (name.hashCode() == hash)
                    endStrings1.add(name);
        }

        // set end elements - reverse the position since we determined the
        //					  not-in-the-other-set-values
        entry.put("domain_0_end", endStrings1);
        entry.put("domain_1_end", endStrings0);

        // only one bottom/top element -> relation is linear
        if ( endStrings1.size() == 1  && endStrings0.size() == 1 )
            entry.put("isLinear", true);
    }

    /**
     * This method looks for the order of relations.
     * That means:
     * 	- hasQuasiOrder   -> isReflexive     & isTransitive
     *  - hasParitalOrder -> hasQuasiOrder   & isAntisymmetric
     *  - hasTotalOrder   -> hasPartialOrder & isLinear
     *  - hasStrictOrder  -> isIrreflexive   & isTransitive
     *
     * @param info
     */
    private void findOrderedRelation(KBEntry entry) {

        if ( entry.has("isReflexive") )  {
            entry.put("hasQuasiOrder", true);

            if ( entry.has("isAntisymmetric") ) {
                entry.put("hasPartialOrder", true);

                if ( entry.has("isLinear") )
                    entry.put("hasTotalOrder", true);
            }

        } else if (entry.has("isIrreflexive") ) {
            if ( entry.has("isAsymmetric") )
                entry.put("hasStrictOrder", true);
        }
    }

}
