package org.eclipse.palamedes.kif.core.knowledgebase.analyser;

import java.util.ArrayList;
import java.util.HashSet;

import org.eclipse.palamedes.kif.core.ast.ASTNode;
import org.eclipse.palamedes.kif.core.ast.KIFObjConst;
import org.eclipse.palamedes.kif.core.ast.KIFRelSent;
import org.eclipse.palamedes.kif.core.ast.KIFSent;
import org.eclipse.palamedes.kif.core.ast.visitor.GroundChecker;
import org.eclipse.palamedes.kif.core.knowledgebase.IKBAnalyser;
import org.eclipse.palamedes.kif.core.knowledgebase.KBEntry;
import org.eclipse.palamedes.kif.core.knowledgebase.KnowledgeBase;

public class ADomainInformation implements IKBAnalyser {

	static final GroundChecker checker  = new GroundChecker();

    public void analyse(KnowledgeBase kb) {
        for (KBEntry entry : kb.getEntries()) {
            isFact(entry);
            determineFactDomainsByPosition(entry);
        }
    }

    /**
     * This method looks if a constant is a fact.
     *
     * This means that top level statements exist and they are ground.
     *
     * Attributes creates: isFact
     *
     * @param info
     */
    private void isFact(KBEntry entry) {

        if ( !entry.getTopLevelNodes().isEmpty() ) {
            for (ASTNode node : entry.getTopLevelNodes() )
                if ( !checker.isGround(node) ) 
                	return;
        	entry.put("isFact", true);
        }
    }

    /**
     * This method determines the domain objects by position.
     *
     * Attributes creates: domain_#, domain_distinct
     * Attributes uses: isFact, hasMultipleArity(optional)
     * Complexity: O(n*a) with n == factCount, a == arity
     *
     */
    private void determineFactDomainsByPosition(KBEntry entry) {

        // only care for facts and do not care about multi arity stuff
        if ( !entry.has("isFact") ||
              entry.has("hasMultipleArity") )
            return;

        // determine arity and do the stuff per positional argument
        ArrayList<KIFSent> relationArray = entry.getTopLevelNodes();
        int 			   nodeCount 	 = relationArray.size();
        String arityString = (String) entry.get(KBEntry.P_ARITY);
        int    arity       = Integer.parseInt(arityString);

        HashSet<String> distinctSet = new HashSet<String>();

        // go over every position
        KIFObjConst objConst;
        for (int i = 0; i < arity; i++) {
            ArrayList<KIFObjConst> objList = new ArrayList<KIFObjConst>();

            // go over every top level node
            for (int j = 0; j < nodeCount; j++) {
                KIFRelSent relation = (KIFRelSent)  relationArray.get(j);
                if ( relation.getTerms().get(i) instanceof KIFObjConst ) {
                    objConst = (KIFObjConst) relation.getTerms().get(i);

                    // add to positional domain
                    objList.add( objConst );

                    // add to distinct elements if not already in
                    String name = objConst.getName().toLowerCase();
                    if ( !distinctSet.contains(name) )
                        distinctSet.add(name);
                }

                // so far if we hit something else then a KIFObjConst we leave
                // here
                // TODO: Check if we could use the information even if we got a
                //       KIFFunTerm, e.g. init relation with cell node
                else
                    return;
            }

            entry.put("domain_" + i, objList);
        }

        entry.put("domain_distinct", distinctSet);
    }

}
